/* 
   Matrix multiplication example

   Pthread bag of tasks version, fine grained solution

   Jim Teresco, CS 338, Williams College
   Mon Feb 17 16:04:31 EST 2003

   $Id: matmult_bagoftasks.c,v 1.1 2003/02/19 20:29:36 terescoj Exp $
*/

/* header files needed for printf, gettimeofday, struct timeval */
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <pthread.h>

/* header file for our own timer.c function diffgettime */
#include "timer.h"

/* we will multiply square matrices, how big? */
#define SIZE 750

/* our matrices - global because of limited stack space in Solaris */
double a[SIZE][SIZE], b[SIZE][SIZE], c[SIZE][SIZE]; 

/* a counter to determine the next available task, and mutex to
   protect it */
int next_avail_task=0;
pthread_mutex_t mutex;

/* function to compute the dot product of row row of a with col col of
   b, storing the result in c[row][col] */
void do_entry(int row, int col) {
  int k;

  /* initialize entry */
  c[row][col] = 0;
  
  /* perform dot product */
  for(k=0; k<SIZE; k++) {
    c[row][col] = c[row][col] + a[row][k]*b[k][col];
  }
}

void *worker(void *arg) {
  int current_task;

  /* lock next_avail_task */
  pthread_mutex_lock(&mutex);
  current_task = next_avail_task;
  next_avail_task++;
  /* unlock next_avail_task */
  pthread_mutex_unlock(&mutex);

  while (current_task < SIZE*SIZE) {
    /* task "current_task" indicates an entry number, counted in
       row major order */
    do_entry(current_task/SIZE, current_task%SIZE);

    /* lock next_avail_task */
    pthread_mutex_lock(&mutex);
    current_task = next_avail_task;
    next_avail_task++;
    /* unlock next_avail_task */
    pthread_mutex_unlock(&mutex);
  }
}

/* it's a simple program for now, we'll just put everything in main */
int main(int argc, char *argv[]) {
  double sum;

  /* counters */
  int i, j;

  /* to pass to gettimeofday to get wall clock times */
  struct timeval start, stop;

  int rc;

  /* pthread stuff */
  int numworkers;
  pthread_t *thread_ids;
  pthread_attr_t attr;

  /* initialize matrices, just fill with junk and do this in serial */
  gettimeofday(&start, NULL);
  for (i=0; i<SIZE; i++) {
    for (j=0; j<SIZE; j++) {
      a[i][j] = i+j;
      b[i][j] = i-j;
    }
  }
  gettimeofday(&stop, NULL);
  printf("Initialization (serial) took: %f seconds\n", diffgettime(start,stop));
  
  /* number of workers is specified on the command-line */
  numworkers=1;
  if (argc > 1) numworkers=atoi(argv[1]);

  thread_ids = (pthread_t *)malloc(numworkers*sizeof(pthread_t));

  /* some pthread setup */
  pthread_setconcurrency(numworkers+1);
  pthread_attr_init(&attr);
  pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);

  /* we'll need a mutex for concurrent access to next_avail_task */
  pthread_mutex_init(&mutex, NULL);

  gettimeofday(&start, NULL);
  /* start workers */
  for (i=0; i<numworkers; i++) {
    rc=pthread_create(&(thread_ids[i]), &attr, &worker, NULL);
    if (rc != 0) {
      fprintf(stderr, "Could not create thread %d\n", i);
      exit(1);
    }
  }

  /* workers working.... */

  /* wait for workers to finish */
  for (i=0; i<numworkers; i++) {
    pthread_join(thread_ids[i], NULL);
  }

  gettimeofday(&stop, NULL);
  printf("Multiplication took: %f seconds\n", diffgettime(start,stop));

  /* This is here to make sure the optimizing compiler doesn't
     get any big ideas about "optimizing" code away completely */
  sum=0;
  for (i=0; i<SIZE; i++) {
    for (j=0; j<SIZE; j++) {
      sum += c[i][j];
    }
  }
  printf("Sum of elements of c=%f\n", sum);

  pthread_mutex_destroy(&mutex);
  free(thread_ids);
 
  return 0;
}
